/* 
 * Connection Guard Module for Apache 2
 * 
 * Copyright 2007 Alexey V. Degtyarev
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: mod_cguard.c,v 1.3 2007/03/05 11:16:03 alexey Exp $
 *
 */

#define MODULE_NAME         "cguard_module"
#define MODULE_VERSION      "0.1"
#define CGUARD_DEBUG_OFF

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_main.h"
#include "ap_mpm.h"
#include "http_log.h"
#include "scoreboard.h"
#include "apr_strmatch.h"

module AP_MODULE_DECLARE_DATA cguard_module;

/* Configuration structure:
 * 
 * enabled                - engine is active for this server;
 * server_max_connects    - number of total simultaneous connections
 *                          to server from ALL clients;
 * server_max_clients     - number of total clients having active
 *                          connection(s) to server at the same time;
 * client_max_connects    - number of total simultaneous connections
 *                          to server from ONE client;
 * request_max_pumps      - limit of simultaneous requests from ONE
 *                          client to ONE URI within server;
 * check_args             - turn on or off checking for arguments to
 *                          exists in client's request;
 * log_enabled            - enable or disable logging process.
 */
typedef struct {
    int enabled;
    int server_max_connects;
    int server_max_clients;
    int client_max_connects;
    int request_max_pumps;
    int check_args;
    int log_enabled;
} cguard_conf;

static int
    mpm_max_clients,                    /* The MaxClients value;        */
    mpm_can_threads,                    /* Doest MPM accept threads?    */
    mpm_server_limit,                   /* ServerLimit for MPM;         */
    mpm_thread_limit;                   /* Threads per process, if MPM  */
                                        /* can do threads, overwise - 1 */
                

#define UNSET -1
                                        /* Default values:             */
#define DEFAULT_SERVER_CONNECTS 32      /*  for server_max_connects    */
#define DEFAULT_SERVER_CLIENTS  16      /*  for server_max_clients     */
#define DEFAULT_REQUEST_PUMPS    8      /*  for pump limit             */

/* Configuration initialization */
static void *create_srv_config(apr_pool_t *pool, server_rec *server)
{
    cguard_conf* conf = apr_pcalloc(pool, sizeof(cguard_conf));

    mpm_max_clients               = DEFAULT_SERVER_CONNECTS;

    conf->enabled                 = UNSET;
    conf->log_enabled             = 1;
    conf->server_max_connects     = DEFAULT_SERVER_CONNECTS;
    conf->client_max_connects     = UNSET;
    conf->request_max_pumps       = DEFAULT_REQUEST_PUMPS;
    conf->check_args              = 0;
    conf->server_max_clients      = UNSET;
    
    return conf;
}

/* Merge configurations */
static void *
merge_srv_conf(apr_pool_t* pool, void* base, void* add)
{
    cguard_conf* old  = base;
    cguard_conf* new  = add;
    cguard_conf* conf = apr_palloc(pool, sizeof(cguard_conf));

    memcpy(conf, old, sizeof(cguard_conf));
    
    if (new->enabled != UNSET)
        conf->enabled = new->enabled;

    if (new->server_max_connects)
        conf->server_max_connects = new->server_max_connects;

    if (new->client_max_connects)
        conf->client_max_connects = new->client_max_connects;

    if (new->server_max_clients) 
        conf->server_max_clients = new->server_max_clients;

    if (new->request_max_pumps)
        conf->request_max_pumps = new->request_max_pumps;

    if (new->log_enabled != UNSET)
        conf->log_enabled = new->log_enabled;

    return conf;
}

static void * 
merge_dir_conf(apr_pool_t* pool, void* base, void* add)
{
    cguard_conf* old  = base;
    cguard_conf* new  = add;
    cguard_conf* conf = apr_palloc(pool, sizeof(cguard_conf));

    memcpy(conf, old, sizeof(cguard_conf));
    
    conf->log_enabled = (new->log_enabled == UNSET) ? 
        old->log_enabled : new->log_enabled;

    return conf;
}

/* CGuardEngine directive handler */
static const char *
set_engine(cmd_parms *cmd, void *dir_cfg, int flag)
{
    /* --------
     * The ServerType directive no longer exists. The method used to serve 
     * requests is now determined by the selection of MPM. There is currently 
     * no MPM designed to be launched by inetd.
     * --------
     *
     *  won't use this:
     * 
     * if (!ap_exists_scoreboard_image())
     */

    if (!ap_extended_status)
        return "mod_cguard: ExtendedStatus must be set to On for CGuardEngine to be enabled";

    cguard_conf* conf = ap_get_module_config(cmd->server->module_config,
            &cguard_module);
    
    if (conf->server_max_clients == UNSET)
        conf->server_max_clients = DEFAULT_SERVER_CLIENTS;

    if (conf->client_max_connects == UNSET)
        conf->client_max_connects = 
            conf->server_max_connects / conf->server_max_clients;

    conf->enabled = flag;
    
    return NULL;
}

/* CGuardLog directive handler */
static const char *
set_log(cmd_parms *cmd, void *dir_cfg, int flag)
{
    cguard_conf* conf = ap_get_module_config(cmd->server->module_config,
            &cguard_module);

    conf->log_enabled = flag;

    return NULL;
}


/* RequestMaxPumps directive handler */
static const char *
set_request_max_pumps(cmd_parms *cmd, void *dir_cfg, const char *arg)
{
    int val;
    
    val = atoi(arg);
    
    /* correct value given? */
    if (val < 1) 
        return "RequestMaxPumps must be an integer greater than 0";

    cguard_conf* conf = ap_get_module_config(cmd->server->module_config,
            &cguard_module);
    
    if (conf->server_max_clients == UNSET)
        conf->server_max_clients = DEFAULT_SERVER_CLIENTS;

    if (conf->client_max_connects == UNSET)
        conf->client_max_connects = 
            conf->server_max_connects / conf->server_max_clients;
    
    if (val > (conf->server_max_connects / conf->server_max_clients)) {
        ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
                "RequestMaxPumps must not exceed value of "
                "(ServerMaxConnects / ServerMaxClients), %d / %d = %d;",
                conf->server_max_connects, conf->server_max_clients,
                conf->server_max_connects / conf->server_max_clients);
        return "Please fix it.";
    }

    if (val > conf->client_max_connects) {
        ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
                "RequestMaxPumps of %d must not exceed value of ClientMaxConnects = %d",
                val, conf->client_max_connects);
        return "Please fix it.";
    }
    
    conf->request_max_pumps = val;
    
    return NULL;
}

/* ServerMaxClients directive handler */
static const char *
set_server_max_clients(cmd_parms *cmd, void *dir_cfg, const char *arg)
{
    int val;

    val = atoi(arg);

    if (val < 1)
        return "ServerMaxClients must be an integer greater than 0";

    cguard_conf* conf = ap_get_module_config(cmd->server->module_config,
            &cguard_module);
    
    if (conf->client_max_connects == UNSET)
        conf->client_max_connects == DEFAULT_REQUEST_PUMPS;
    
    if (val > (conf->server_max_connects / conf->client_max_connects)) {
        ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
                "ServerMaxClients of %d must not exceed value of "
                "(ServerMaxConnects / ClientMaxConnects), %d / %d = %d;",
                val, conf->server_max_connects, conf->client_max_connects,
                conf->server_max_connects / conf->client_max_connects);
        return "Please fix it.";
    }

    conf->server_max_clients = val;
    
    return NULL;
}

/* ClientMaxConnects directive handler */
static const char *
set_client_max_connects(cmd_parms *cmd, void *dir_cfg, const char *arg)
{
    int val;

    val = atoi(arg);

    if (val < 1)
        return "ClientMaxConnects must be an integer greater than 0";

    cguard_conf* conf = ap_get_module_config(cmd->server->module_config,
            &cguard_module);
    
    if (conf->server_max_clients == UNSET)
        return "ServerMaxClients must be set before ClientMaxConnects";

    if (val > conf->server_max_connects / conf->server_max_clients) {
        ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
                "ClientMaxConnects of %d must not exceed value of "
                "(ServerMaxConnects / ServerMaxClients), %d / %d = %d;",
                val, conf->server_max_connects, conf->server_max_clients,
                conf->server_max_connects / conf->server_max_clients);
        return "Please fix it.";
    }

    conf->client_max_connects = val;
    
    return NULL;
}

/* ServerMaxConnects directive handler */
static const char *
set_server_max_connects(cmd_parms *cmd, void *dir_cfg, const char *arg)
{
    int val;
    val = atoi(arg);

    if (val > mpm_max_clients) {
        ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
                "ServerMaxConnects of %d must not exceed value of MaxClients = %d",
                val, mpm_max_clients);
        return "Please fix it.";
    }

    cguard_conf* conf = ap_get_module_config(cmd->server->module_config,
                &cguard_module);
    
    conf->server_max_connects = val;

    return NULL;
}

/* PumpCheckArgs directive handler */
static const char
*set_pump_check_args(cmd_parms *cmd, void *dir_cfg, int flag)
{
    cguard_conf* conf = ap_get_module_config(cmd->server->module_config,
            &cguard_module);
    
    conf->check_args = flag;
    
    return NULL;
}

/* MaxClients directive handler */
static const char 
*set_max_clients(cmd_parms *cmd, void *dir_cfg, const char *arg)
{
    mpm_max_clients = atoi(arg);
    
    cguard_conf* conf = ap_get_module_config(cmd->server->module_config,
            &cguard_module);

    conf->server_max_connects = mpm_max_clients;

    return NULL;
}

/* ExtendedStatus directive handler. AS IS from  mod_status.c */
static const char *set_extended_status(cmd_parms *cmd, void *dummy, int arg)
{
    const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
    if (err != NULL) {
        return err;
    }
    ap_extended_status = arg;
    return NULL;    
}

/* module directives */
static const command_rec cguard_cmds[] = {

    /* Third-party modules directives.  They are need to set up correct values
     * of some variables of our module */
    AP_INIT_TAKE1("MaxClients", set_max_clients, NULL, RSRC_CONF,
            "Maximum value of MaxClients for this run of Apache"),

    AP_INIT_FLAG("ExtendedStatus", set_extended_status, NULL, RSRC_CONF,
               "\"On\" to enable extended status information, \"Off\" to disable"),

    /* this module directives itself: */
    
    /* Turn on or off the limiting engine. */
    AP_INIT_FLAG("CGuardEngine", set_engine, NULL, RSRC_CONF,
            "Enables or disables the limiting engine"),

    /* Log if any limiting parameter are trying to be overquotted. */
    AP_INIT_FLAG("CGuardLog", set_log, NULL, OR_ALL,
            "\"On\" to enable log messages in ErrorLog about one of the catched exceeding"),
    
    /* Total number of server's connections. */
    AP_INIT_TAKE1("ServerMaxConnects", set_server_max_connects, NULL, RSRC_CONF,
            "Limit of all connections to the server"),

    /* Total number of clients, simultaneously connected to the server. */
    AP_INIT_TAKE1("ServerMaxClients", set_server_max_clients, NULL, RSRC_CONF,
            "Integer non-negative limit of simultaneous connections to the server from one IP"),
    
    /* Connections for one remote client. Note, "remote client" in this module
     * understanding - is not the same as in apache's meaning. The very simple example
     * describes this diffrence.
     *   imagine situation:
     *    1.2.3.4  GET /foo 
     *    1.2.3.5  GET /bar
     *    1.2.3.5  GET /bar?foo
     *    1.2.3.5  GET /bar
     *  ---------
     *       apache's meaning: 4 requests, 4 remote clients
     *   mod_cguard's meaning: 4 requests, 2 remote clients
     */     
    AP_INIT_TAKE1("ClientMaxConnects", set_client_max_connects, NULL, RSRC_CONF,
            "Limit of simultaneous connections to server for one client"),
    
    /* Stream limit for one client. No more simultaneous connections to THE SAME
     * URI within one server. */
    AP_INIT_TAKE1("RequestMaxPumps", set_request_max_pumps, NULL, RSRC_CONF,
            "Integer non-negative limit of simultaneous connections, requesting the same resource"),

    /* Flag showing if mod_cguard should check the arguments in the client's 
     * request while counting pump limit. In the example above there are:
     * one request from 1.2.3.4 and three ones from 1.2.3.5. With
     * "PumpCheckArgs On":
     *   /bar     x 2 requests
     *   /bar?foo x 1 request
     *       ---------
     *              3 requests
     * Imagine RequestMaxPumps set to 3, client are able to ask for
     *   /bar     x 1 request
     *   /bar?foo x 2 requests
     *
     * and with default value of PumpCheckArgs (Off) there are no any available
     * requests for /bar with or without any params in query.    
     */
    AP_INIT_FLAG("PumpCheckArgs", set_pump_check_args, NULL, OR_ALL,
            "Set to YES to enable strict checks while comparing requests"),
    {NULL}
};

/* 
 * Log routine 
 * 
 * text  - string information for log;
 * debug - flag for "debug" information.
 */
static void cguard_log(request_rec *r, int debug, const char *text, ...)
{
    char str[512];
    va_list ap;
    
    cguard_conf* conf = ap_get_module_config(r->server->module_config,
            &cguard_module);
    
    va_start(ap, text);
    apr_vsnprintf(str, sizeof(str), text, ap);

    if (debug == 1) {
#ifdef CGUARD_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "[cguard debug] %s", str);
#endif
    }
    else if (conf->log_enabled)
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "%s", str);
    
    va_end(ap);
    
    return;
}

/* Module initialization */
static int cguard_init(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp,
        server_rec *s)
{
    ap_mpm_query(AP_MPMQ_IS_THREADED, &mpm_can_threads);
    
    if (mpm_can_threads > 0)
        ap_mpm_query(AP_MPMQ_MAX_THREADS, &mpm_thread_limit);
    else mpm_thread_limit = 1;

    ap_mpm_query(AP_MPMQ_MAX_DAEMONS, &mpm_server_limit);
    return NULL;
}

/* Main handler */
static int cguard_handler(request_rec *r)
{
    int j, i,                     
        server_connects    = 0,     
        server_clients    = 0,
        client_connects    = 1,
           request_pumps    = 0;
                        
    const char 
        *server_name,
        *remote_ip;

    const apr_strmatch_pattern *pattern;

    apr_hash_t *clients = NULL;
    clients = apr_hash_make(r->pool);
    
    worker_score  *ws_record;
    process_score *ps_record;
    
    cguard_conf* conf = ap_get_module_config(r->server->module_config,
            &cguard_module);
    
#ifdef CGUARD_DEBUG
    cguard_log(r, 1, "Server %s: Engine %s, MaxConnects=%d, "
            "MaxClients=%d, ClientQuota=%d, PumpLimit=%d",
            ap_get_server_name(r), (conf->enabled ? "Enabled" : "Disabled"),
            conf->server_max_connects, conf->server_max_clients,
            conf->client_max_connects, conf->request_max_pumps);
#endif
    
    if (!ap_is_initial_req(r))
        return DECLINED;
    
    if (!conf->enabled || conf->enabled == UNSET)
        return DECLINED;

    server_name = ap_get_server_name(r);
    
    for (i = 0; i < mpm_server_limit; ++i) {
        for (j = 0; j < mpm_thread_limit; ++j) {
            ps_record = ap_get_scoreboard_process(i);
            ws_record = ap_get_scoreboard_worker(i, j);
            
            if    /* We should count all BUSY-statuses + GRACEFUL, except KEEPALIVE */
              (!(  ws_record->status == SERVER_BUSY_READ 
                || ws_record->status == SERVER_BUSY_WRITE
                || ws_record->status == SERVER_BUSY_LOG
                || ws_record->status == SERVER_BUSY_DNS
            /*    || ws_record->status == SERVER_BUSY_KEEPALIVE */
                || ws_record->status == SERVER_GRACEFUL
              ))
                  continue;
            
            remote_ip = ws_record->client;
            if (apr_hash_get(clients, remote_ip, APR_HASH_KEY_STRING) == NULL) {
                apr_hash_set(clients, remote_ip, APR_HASH_KEY_STRING, "-");
                server_clients++;
            }
            else if (!strcmp(r->connection->remote_ip, remote_ip))
                client_connects++;
        
            if (strcmp(server_name, ws_record->vhost)) 
                continue;
            
            server_connects++;
                
            if (strcmp(r->connection->remote_ip, remote_ip))
                continue;
            
            pattern = apr_strmatch_precompile(r->pool, conf->check_args == 1 ? 
                    r->unparsed_uri : r->uri, 0);
            if (apr_strmatch(pattern, ws_record->request,
                        strlen(ws_record->request)) != NULL)
                request_pumps++;
        }
    }

    if (server_connects > conf->server_max_connects) {
        cguard_log(r, 0, "Access to %s denied - ServerMaxConnects reached "
                "(%d allowed, %d currently)",
                server_name, conf->server_max_connects, client_connects);
        return HTTP_SERVICE_UNAVAILABLE;
    }

    if (server_clients > conf->server_max_clients) {
        cguard_log(r, 0, "Access to %s denied - ServerMaxClients reached "
                "(%d allowed, %d currently)",
                server_name, conf->server_max_clients, server_clients);
        return HTTP_SERVICE_UNAVAILABLE;
    }

    if (client_connects > conf->client_max_connects) {
        cguard_log(r, 0, "Access to %s denied - ClientMaxConnects reached "
                "(%d allowed, %d currently)",
                server_name, conf->client_max_connects, client_connects);
        return HTTP_SERVICE_UNAVAILABLE;
    }

    if (request_pumps > conf->request_max_pumps) {
        cguard_log(r, 0, "Access to %s denied - RequestMaxPumps reached "
                "(%d allowed, %d currently)",
                r->uri, conf->request_max_pumps, request_pumps);
        return HTTP_SERVICE_UNAVAILABLE;
    }

#ifdef CGUARD_DEBUG
    cguard_log(r, 1, "Server: Connects=%d, Clients=%d",
            server_connects, server_clients);
    cguard_log(r, 1, "Client: IP=%s, Connects=%d",
            r->connection->remote_ip, client_connects, request_pumps);
    cguard_log(r, 1, "Request: Hostname=%s, uri=[%s], base=[%s], "
            "file=[%s], Pumps=%d",
            r->hostname, r->unparsed_uri, r->uri,
            r->filename, request_pumps);
#endif
    
    return DECLINED;
}

/* hooks */
static void register_hooks(apr_pool_t *pool)
{
    ap_hook_post_config(cguard_init, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_access_checker(cguard_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

/* Structure of module in Apache 2 */
module AP_MODULE_DECLARE_DATA cguard_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,                       /* dir config creater */
    merge_dir_conf,             /* dir merger */
    create_srv_config,          /* server config */
    merge_srv_conf,             /* merge server config */
    cguard_cmds,                /* command table */
    register_hooks              /* register_hooks */
};
